package jfftw;
import java.io.File;

/**       Class to implement simple one-off Real to Complex and Complex to
 *        RealFourier transforms using fftw as a native library.
 *        <p>
 *        A new plan is created for each FFT. 
 *        This is sub-optimal for a large number of
 *        FFTs, but reasonable for ``one-off''s.
 *        @author Will Hossack, The University of Edinburgh
 *        @version 1.2
 *
 */
public class FFTWReal extends FFTW {

    /**      Default constructor.
     */
    public FFTWReal() {
    }


    /**      Constructor to for a FFTWReal from a FFTW, 
     *       The PlanFlag is also copied.
     *       @param fft the FFTW 
     *       @since 1.1
     */
    public FFTWReal(FFTW fft) {
	super(fft);
    }

    /**      Constructor to optionally load system wisdom file if available.
     *       @param systemWisdom if true loads system wisdom file.
     */
    public FFTWReal(boolean systemWisdom) {
	super(systemWisdom);
    }

    /**     Constructor to load specified wisdom file is available.
     *      @param wisdomFile
     */
    public FFTWReal(File wisdomFile) {
	super(wisdomFile);
    }

    /**      Constructor to load specified wisdom file by name.
     *       @param wisdomFile
     */
    public FFTWReal(String wisdomFile) {
	super(wisdomFile);
    }


    /**     Method to take a one dimensional Forward FFT of
     *      a real data array of n elements. 
     *      <p>
     *      The output is Complex, held in a <code>double[]</code> 
     *      array with alternative 
     *      real/imaginary parts with a total of <code>n/2+1</code> complex pairs.
     *      <p>
     *      Note the input and output arrays are of different length.
     *      <p>
     *      This method implements out-of-place transforms only, and the
     *      input real array is not modified.
     *      @param realArray the input real array (not changed)
     *      @return double[] holding the FFT in n/2+1 complex pairs.
     */
    public double[] oneDimensionalForward(double realArray[]) {

	if (realArray == null || realArray.length == 0) {
	    throw new NullPointerException("FFTWReal.oneDimensionalForward: called with null or zero length data array.");
	}

	//        Make output array of the correct length
	double complexArray[] = new double[2*(realArray.length/2 + 1)];
	
	if (loaded) {
	    nativeOneDimensionalForward(realArray,complexArray,getPlanFlag());
	}
	else {
	    errorStream.println("FFTWReal.oneDimensionalForward: library not loaded.");
	}
	return complexArray;
    }

    /**       Private native implement the oneDimensionalForward
     */
    private native void nativeOneDimensionalForward(double in[],
						    double out[],
						    int flag);

    /**     Method to take a one dimensional Backward FFT of
     *      a complex hermition array to give Real output.
     *      Th Complex input is held  as double array with alternative 
     *      real/imaginary parts with a total of <code>n/2+1</code> Complex pairs.
     *      <p>
     *      The output is a double array of <code>n</code> elements;
     *      note the input and output arrays are of different length.
     *      <p>
     *      This method implements out-of-place transforms.
     *      @param complexArray array of <code>n/2+1</code> complex parts
     *      @return double[] real array of n real values
     */
    public double[] oneDimensionalBackward(double complexArray[]) {
	
	if (complexArray == null || complexArray.length == 0) {
	    throw new NullPointerException("FFTWReal.oneDimensionalBackward: called with null or zero length data array.");
	}
	//         Make real array of the correct length
	double realArray[] = new double[2*(complexArray.length/2 - 1)];

	if (loaded) {
	    nativeOneDimensionalBackward(complexArray,realArray,getPlanFlag());
	}
	else {
	   errorStream.println("FFTWReal.oneDimensionalBackward: library not loaded.");
	} 
	return realArray;
    }

    /**       Private native implement the oneDimensionalForward
     */
    private native void nativeOneDimensionalBackward(double in[],
						    double out[],
						     int flag);


    /**    Method to take the two-dimensional Forward FFT of a real data
     *     held in one dimensional double array.
     *     the <code>i,j</code> pixel of the real image is located in array element </code>j*width + i</code>.
     *     <p>
     *     The Complex FFT is returned in a double array with real/imag parts in 
     *     even/odd elements, with the <code>k,l</code> Complex components in
     *      <p>
     *       Real part in location <code>2(l x wft + k)</code><br>
     *       Imag part in location <code>2(l x wft + k) + 1</code><br>
     *     where <code>wft = width/2 + 1</code>
     *     @param width the image width.
     *     @param height the image height.
     *     @param realArray the image data in <code>double[]</code> of length <code>width*height</code>.
     *     @return <code>double[]</code> DFT packed into one dimensional array
     *     @throws <code>IllegalArgumentException</code> if <code>width*height</code> does not match the array length. 
     *     
     */
    public double[] twoDimensionalForward(int width, int height, double realArray[]) {

	//            Basic sanity checks.... 
	if (realArray == null || realArray.length == 0) {
	    throw new NullPointerException("FFTWReal.twoDimensionalForward: called with null or zero length data array.");
	}
	if (realArray.length != width*height) {
	    throw new IllegalArgumentException("FFTWReal.twoDimensionalForward: called with width: " + width +
					       " height: " + height + " but array length of: " + realArray.length);
	}

	int number = 2*height*(width/2 + 1);              // Length of complex array
	double complexArray[] = new double[number];       // Make output array
	if (loaded) {
	    nativeTwoDimensionalForward(width,height,realArray,complexArray,getPlanFlag());
	}
	else {
	    errorStream.println("FFTWReal.twoDimensionalForward: library not loaded.");
	} 
	return complexArray;
    }


    /**       Private native implement the twoDimensionalForward
     */
    private native void nativeTwoDimensionalForward(int width, int height, double realArray[],
						    double complexArray[], int flag);



    /**    Method to take the two dimensional backward DFT of a hermition 
     *     complex DFT held in a double array with real/imag parts in 
     *     even/odd elements, with the <code>k,l</code> Complex components in     
     *     <p>
     *       Real part located at <code>2(j * wft + i)</code><br>
     *       Imag part located at <code>2(j * wft + i) + 1</code><br>
     *     where <code>wft = width/2 + 1</code>.
     *     <p>
     *     The transformed real data is returned in a one-dimension double array
     *     with pixel <code>i,j</code> located at element <code>j*width + i</code>.
     *
     *     @param width the image width
     *     @param height the image height
     *     @param complexArray the Fourier data packed in a 1-d array.
     *     @return <code>double[]</code> DFT packed into one dimensional array
     *     @throws <code>IllegalArgumentException</code> if width*height is not consistent with array length.
     */
    public double[] twoDimensionalBackward(int width, int height, double complexArray[]) {

		//            Basic sanity checks.... 
	if (complexArray == null || complexArray.length == 0) {
	    throw new NullPointerException("FFTWReal.twoDimensionalBackward: called with null or zero length data array.");
	}
	if (complexArray.length != 2*height*(width/2 + 1)) {
	    throw new IllegalArgumentException("FFTWReal.twoDimensionalBackward: called with width: " + width +
					       " height: " + height + " but array length of: " + complexArray.length);
	}
	double realArray[] = new double[width*height];        // Make output array
	if (loaded) {
	    nativeTwoDimensionalBackward(width,height,complexArray,realArray,getPlanFlag());
	}
	else {
	    errorStream.println("FFTWReal.twoDimensionalBackward: library not loaded.");
	} 
	return realArray;
    }


    
    /**       Private native implement the twoDimensionalBackward
     */
    private native void nativeTwoDimensionalBackward(int width, int height, 
						     double complexArray[],
						     double realArray[], int flag);



    /**    Method to take the three dimensional Forward DFT of a real cube
     *     held in one dimensional double array.
     *     the <code>i,j,k</code> pixel of the real cube is located in array element 
     *     <code>k * width * height + j * width + i</code>
     *     <p>
     *     The Complex DFT is returned in a double array with real/imag parts in 
     *     even/odd elements, with the i,j Complex components in
:    *     <p>
     *       Real part located at <code>2(k * wft * height + j * wft + i)</code><br>
     *       Imag part located at <code>2(k * wft * height + j * wft + i)</code> + 1<br>
     *       where <code>wft = width/2 + 1</code>.
     *
     *     @param width the cube width
     *     @param height the cube height
     *     @param depth the cube depth
     *     @param realArray the cube data in double[] of length width*height*depth
     *     @return <code>double[]</code> DFT packed into one dimensional array
     *     @throws <code>IllegalArgumentException</code> if <code>width*height*depth</code> is not equal the array length.
     *     
     */
    public double[] threeDimensionalForward(int width, int height, int depth, 
					    double realArray[]) {

		//            Basic sanity checks.... 
	if (realArray == null || realArray.length == 0) {
	    throw new NullPointerException("FFTWReal.threeDimensionalForward: called with null or zero length data array.");
	}
	if (realArray.length != width*height*depth) {
	    throw new IllegalArgumentException("FFTWReal.threeDimensionalForward: called with width: " + width +
					       " height: " + height + " depth : " + depth + 
					       " but array length of: " + realArray.length);
	}


	int number = 2*depth*height*(width/2 + 1);        // length of output array
	double complexArray[] = new double[number];       // Make output array
	if (loaded) {
	    nativeThreeDimensionalForward(width,height,depth,realArray,
					  complexArray,getPlanFlag());
	}
	else {
	    errorStream.println("FFTWReal.threeDimensionalForward: library not loaded.");
	} 
	    
	return complexArray;
    }

    /**   Private native method to implement the three-dimensional FFT
     */
    private native void nativeThreeDimensionalForward(int width, int height, int depth, 
						      double realArray[], 
						      double complexArray[],
						      int flag);


    /**    Method to take the three dimensional backward DFT of a hermition 
     *     complex DFT held in a double array with real/imag parts in 
     *     even/odd elements, with the i,j Complex components in
:    *     <p>
     *       Real part located at  <code>2(k x wft x height + j x wft + i)</code><br>
     *       Imag part located at  <code>2(k x wft x height + j x wft + i) + 1</code><br>
     *       where <code>wft = width/2 + 1</code>.
     *     <p>
     *     The transformed real data is returned in a one-dimensional double array
     *     with pixel i,j located at element <code>k x width x height + j x width + i</code>.
     *
     *     @param width the cube width
     *     @param height the cube height
     *     @param depth the cube depth
     *     @param complexArray Fourier data packed in a 1-d array.
     *     @return double[] DFT packed into one dimensional array
     *     @throws <code>IllegalArgumentException</code> if width, height, depth is not consistent with array length.
     */    
    public double[] threeDimensionalBackward(int width, int height, int depth, 
					     double complexArray[]) {

			//            Basic sanity checks.... 
	if (complexArray == null || complexArray.length == 0) {
	    throw new NullPointerException("FFTWReal.threeDimensionalBackward: called with null or zero length data array.");
	}
	if (complexArray.length != 2*depth*height*(width/2 + 1)) {
	    throw new IllegalArgumentException("FFTWReal.threeDimensionalBackword: called with width: " + width +
					       " height: " + height + " depth : " + depth + 
					       " but array length of: " + complexArray.length);
	}


	double realArray[] = new double[depth*height*width];      // Make output array
	if (loaded) {
	    nativeThreeDimensionalBackward(width,height,depth,complexArray,realArray,
					   getPlanFlag());
	}
	else {
	    errorStream.println("FFTWReal.threeDimensionalBackward: library not loaded.");
	} 
	
	return realArray;
    }

    /**   Private native method to implement the three-dimensional FFT
     */
    private native void nativeThreeDimensionalBackward(int width, int height, int depth, 
						       double complexArray[], 
						       double realArray[],
						       int flag);

}
